/*
 * Copyright (c) 2001 Mark Fullmer and The Ohio State University
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *      $Id: fmt.c,v 1.4 2001/02/10 04:40:12 maf Exp $
 */

#include <strings.h>
#include "fmt.h"

unsigned int fmt_uint8(register char *s, register u_int8 u, int format)
{
  register int len = 0;
  char *s1;

  if (!s)
    return 0;

  s1 = s;

  s += FMT_UINT8 - 1;
  do {

    ++len;
    *--s = '0' + (u % 10);
    u /= 10;

  } while(u); 

  if ((format == FMT_PAD_RIGHT) || (format == FMT_JUST_LEFT)) {
    bcopy(s, s1, len);
    if (format == FMT_PAD_RIGHT)
      for (; len < (FMT_UINT8 - 1); ++len)
        s1[len] = ' ';
    s1[len] = 0;
    return len;
  }

  return len;

} /* fmt_uint8 */

unsigned int fmt_uint16(register char *s, register u_int16 u, int format)
{
  register int len = 0;
  char *s1;

  if (!s)
    return 0;

  s1 = s;

  s += FMT_UINT16 - 1;
  do {

    ++len;
    *--s = '0' + (u % 10);
    u /= 10;

  } while(u); 

  if ((format == FMT_PAD_RIGHT) || (format == FMT_JUST_LEFT)) {
    bcopy(s, s1, len);
    if (format == FMT_PAD_RIGHT)
      for (; len < (FMT_UINT16 - 1); ++len)
        s1[len] = ' ';
    s1[len] = 0;
    return len;
  }

  return len;

} /* fmt_uint16 */

unsigned int fmt_uint32(register char *s, register u_int32 u, int format)
{
  register int len = 0;
  char *s1;
  int i;

  if (!s)
    return 0;

  s1 = s;

  s += FMT_UINT32 - 1;
  do {

    ++len;
    *--s = '0' + (u % 10);
    u /= 10;

  } while(u); 

  if ((format == FMT_PAD_RIGHT) || (format == FMT_JUST_LEFT)) {
    bcopy(s, s1, len);
    if (format == FMT_PAD_RIGHT)
      for (; len < (FMT_UINT32 - 1); ++len)
        s1[len] = ' ';
    s1[len] = 0;
    return len;
  }

  if (format == FMT_PAD_LEFT) {
    for (i = 0; i < ((FMT_UINT32 - 1) - len); ++i)
      s1[i] = ' ';
    s1[(FMT_UINT32 - 1)] = 0;
    return (FMT_UINT32 - 1);
  }

  return 0;

} /* fmt_uint32 */

unsigned int fmt_uint64(register char *s, register u_int64 u, int format)
{
  register int len = 0;
  char *s1;
  int i;

  if (!s)
    return 0;

  s1 = s;

  s += FMT_UINT64 - 1;

  do {

    ++len;
    *--s = '0' + (u % 10);
    u /= 10;

  } while(u); 

  if ((format == FMT_PAD_RIGHT) || (format == FMT_JUST_LEFT)) {
    bcopy(s, s1, len);
    if (format == FMT_PAD_RIGHT)
      for (; len < (FMT_UINT64 - 1); ++len)
        s1[len] = ' ';
    s1[len] = 0;
    return len;
  }

  if (format == FMT_PAD_LEFT) {
    for (i = 0; i < ((FMT_UINT64 - 1) - len); ++i)
      s1[i] = ' ';
    s1[(FMT_UINT64 - 1)] = 0;
    return (FMT_UINT64 - 1);
  }

  return 0;

} /* fmt_uint64 */

unsigned int fmt_ipv4(register char *s, register u_int32 u, int format)
{
  register int len = 0;
  char *s1;
  int i, j;
  u_int8 e[4];
  char c[4][4];

  if (!s)
    return 0;

  j = 0;

  e[0] = (u & 0xFF000000)>>24;
  e[1] = (u & 0x00FF0000)>>16;
  e[2] = (u & 0x0000FF00)>>8;
  e[3] = (u & 0x000000FF);

  for (i = 0; i < 4; ++i) {
    s1 = &c[i][3];
    len = 0;
    do {
      ++len;
      *--s1 = '0' + (e[i] % 10);
      e[i] /= 10;
    } while(e[i]); 

    bcopy(s1, s+j, len);
    j += len;
    *(s+j) = '.';
    ++j;
  }

  --j;
  s[j] = 0;

  if (format == FMT_JUST_LEFT)
    return j;

  if (format == FMT_PAD_RIGHT) {
    for (; j < (FMT_IPV4-1); ++j)
      s[j] = ' ';
    s[j] = 0;
    return (FMT_IPV4-1);
  }

  if (format == FMT_JUST_RIGHT) {
    bcopy(s, s+(FMT_IPV4-1)-j, j);
    for (i = 0; i < (FMT_IPV4-1)-j; ++i)
      s[i] = ' ';
    s[FMT_IPV4-1] = 0;
    return (FMT_IPV4-1);
  }

  return j;

} /* fmt_ipv4 */

